/**
 * File:
 *  modbusSlave.c
 *
 * Notes:
 *  This file contains the implementation for a serial RTU modbus slave.
 *
 * Functions:
 *  addModbusBlock  Creates an I/O block of a specified type
 *  calcCRC         Calculates the CRC for the message content
 *  coilState       Sets the state of a coil
 *  decodePacket    Decodes a packet received from modbus master
 *  modbusSlave     Initialise modbus slave structure
 *  packBits        Packs bits into a message buffer
 *  packRegisters   Packs registers into a message buffer
 *  setRegister     Sets the value of a holding register
 *  startTimeout    Starts the message timeout timer
 *  restartRx       Restart communications
 *
 * History:
 *  10/06/2012 Written by Simon Platten
 *  19/06/2012 Added receivePacket function which buffers a packet
 *  21/06/2012 Implemented exception responses
 *             Implemented Timer0 to time receipt of bytes
 *  25/06/2012 Implemented startTimeout
 *  27/06/2012 Replaced individual request and response buffers with a single
 *             message buffer.  Re-wrote CRC to use lookup tables.  Improved
 *             comments.
 *  30/06/2012 Implemented register mappings to allow 8 bit I/O ports to be
 *             mapped into 16 bit registers
 *  02/07/2012 Re-wrote I/O structures and functions, all I/O is now added
 *             using addModbusBlock.
 *  07/07/2012 Removed static variables:
 *              pCoilsTail,
 *              pStatusBitsTail,
 *              pHoldingRegsTail,
 *              pInputRegsTail,
 *              bytStartHi,
 *              bytStartLo, 
 *              bytTotalHi, 
 *              bytTotalLo, 
 *              blnValidStart,
 *              blnValidEnd;
 */
#include <built_in.h>

#include "modbus.h"
// Pointer to the I/O linked lists
static modbusBlockDef* pCoils       = NULL;
static modbusBlockDef* pStatusBits  = NULL;
static modbusBlockDef* pHoldingRegs = NULL;
static modbusBlockDef* pInputRegs   = NULL;
// Exception code
static mbException exceptionCode;
// The modbus slave address
static byte bytSlaveAddress;
// The byte index
static byte bytIndex = 0;
// Receiver & transmitter GAP set-points
static uint uintRxGapSetPt1;
static uint uintRxGapSetPt2;
static uint uintRxGapSetPt3;
static uint uintRxGapSetPt;
// Receiver GAP counter
static uint uintRxGap;
// Received packet buffer
static byte aryBuffer[MAX_PACKET_LENGTH];
static byte bytRxphase;
// Pointer to the last block that was modified
static modbusBlockDef* pWriteBlock;
// RS-485 direction control bit
sbit Tx_dir at LATE.B0;
/**
 * Function:
 *  calcCRC
 *
 * Parameters:
 *  aryBuffer, the message containing the bytes to calculate the CRC for
 *  uintLength, the number of bytes in the message
 *
 * Returns:
 *  A 16bit CRC for the passed message
 *
 * Uses: R0..3, FSR0
 * Remarks: 650 cycles for 8 bytes of data
 */
static uint calcCRC(uint uintLength) {
  asm {
   MBgetCRC:
         movff    FARG_ModbusSlave_calcCRC_uintLength,R2  // byte count in R2
         lfsr     0,ModbusSlave_aryBuffer                 // point to aryBuffer
         setf     R0                // preset CRC
         setf     R1
   MBgetCRC1:
         movf     POSTINC0,W        // xor byte with CRClo
         xorwf    R0,F
         movlw    8                 // bit count in R3
         movwf    R3
   MBgetCRC2:
         bcf      STATUS,C
         rrcf     R1,F
         rrcf     R0,F
         bnc      $+10
         movlw    0xA0
         xorwf    R1,F
         movlw    0x01
         xorwf    R0,F
         decfsz   R3,F
         bra      MBgetCRC2
         decfsz   R2,F
         bra      MBgetCRC1
         return
  }
  return 0;
}
/**
 * Function:
 *  coilState
 *
 * Parameters:
 *  uintAddress, the address of the coil to modify
 *  blnState, TRUE to turn coil on, FALSE to turn coil off
 *
 * Returns:
 *  TRUE if operation completed else FALSE
 */
static boolean coilState(uint uintAddress, boolean blnState) {
// Find a coil block that contains the address
  if ( pWriteBlock == NULL ) {
// No write block set, find it!
    pWriteBlock = pCoils;
    while( pWriteBlock != NULL ) {
      if ( uintAddress >= pWriteBlock->uintAddress
        && uintAddress <= (pWriteBlock->uintAddress + 
                           pWriteBlock->uintTotal - 1)  ) {
        break;
      }
      pWriteBlock = pWriteBlock->pNext;
    }
  }
  if ( pWriteBlock != NULL ) {
// Get byte and bit from address
    byte bytIndex, bytBit;
    uint uintOffset;
    uintOffset = uintAddress - pWriteBlock->uintAddress;
    bytIndex = uintOffset / 8;
    bytBit = 1 << (uintOffset % 8);

    if ( blnState == TRUE ) {
// Turn on
      ((byte*)pWriteBlock->paryData)[bytIndex] |= bytBit;
    } else {
// Turn off
      ((byte*)pWriteBlock->paryData)[bytIndex] &= (0xff ^ bytBit);
    }
    return TRUE;
  }
  return FALSE;
}
/**
 * Function:
 *  packBits
 *
 * Parameters:
 *  eType, should be either COILS or STATUS_INPUTS
 *  pBuffer, a pointer to the buffer to pack the bits into
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  The number of bytes the bits were packed into
 */
static uint packBits(mbType eType, byte* pBuffer,
                     uint uintStart, uint uintEnd) {
// Find the first I/O point
  modbusBlockDef* pNode;
  uint uintBytes = 0;

  if ( eType == COILS ) {
    pNode = pCoils;
  } else if( eType == STATUS_INPUTS ) {
    pNode = pStatusBits;
  } else {
    return 0;
  }
  while( pNode != NULL ) {
    if ( uintStart >= pNode->uintAddress ) {
      break;
    }
    pNode = pNode->pNext;
  }
  if ( pNode != NULL && uintEnd < pNode->uintAddress + pNode->uintTotal ) {
    uint uintOffset;
    byte bytIndex;
// Work out the address offset
    uintOffset = uintStart - pNode->uintAddress;
// The index into the data array
    bytIndex = uintOffset / 8;
// The bit in the data
    Hi(uintOffset) = 1 << (Lo(uintOffset) & 7);
// The bit of response bit field
    Lo(uintOffset) = 1;
// Clear first byte of response bit field
    *pBuffer = 0;
    while( uintStart <= uintEnd ) {
      if ( Lo(uintOffset).B0 ) {
        uintBytes++;
      }
// State
      if ( (((byte*)pNode->paryData)[bytIndex] & Hi(uintOffset)) ) {
        *pBuffer |= Lo(uintOffset);
      }
      Hi(uintOffset) <<= 1;
      
      if ( Hi(uintOffset) == 0 ) {
        Hi(uintOffset) = 1;
        bytIndex++;
      }
      Lo(uintOffset) <<= 1;
      
      if ( Lo(uintOffset) == 0 ) {
        Lo(uintOffset) = 1;
        pBuffer++;
        *pBuffer = 0;
      }
// Next address
      uintStart++;
    }
  }
  return uintBytes;
}
/**
 * Function:
 *  packRegisters
 *
 * Parameters:
 *  eType, should be either HOLDING_REGISTERS or INPUT_REGISTERS
 *  pBuffer, a pointer to the buffer to pack the registers into
 *  uintStart, the start address
 *  uintEnd, the end address
 *
 * Returns:
 *  The number of bytes the registers were packed into
 */
static uint packRegisters(mbType eType, byte* pBuffer,
                          uint uintStart, uint uintEnd) {
// Find the first I/O point
  modbusBlockDef* pNode;
  uint uintBytes = 0;

  if ( eType == HOLDING_REGISTERS ) {
    pNode = pHoldingRegs;
  } else if( eType == INPUT_REGISTERS ) {
    pNode = pInputRegs;
  } else {
    return 0;
  }
  while( pNode != NULL ) {
    if ( uintStart >= pNode->uintAddress ) {
      break;
    }
    pNode = pNode->pNext;
  }
  if ( pNode != NULL && uintEnd < (pNode->uintAddress + pNode->uintTotal) ) {
    uint uintOffset;
    uintOffset = uintStart - pNode->uintAddress;
    while( uintStart <= uintEnd ) {
      pBuffer[uintBytes++] = Hi(((uint*)pNode->paryData)[uintOffset]);
      pBuffer[uintBytes++] = Lo(((uint*)pNode->paryData)[uintOffset]);
// Next address
      uintStart++;
      uintOffset++;
     }
  }
  return uintBytes;
}
/**
 * Function:
 *  setRegister
 *
 * Parameters:
 *  uintAddress, the address of the register to modify
 *  bytDataHi, the hi-byte of the data
 *  bytDataLo, the lo-byte of the data
 *
 * Returns:
 *  TRUE if operation completed else FALSE
 */
static boolean setRegister(uint uintAddress, byte bytDataHi, byte bytDataLo) {
// Find a holding register block that contains the address
  if ( pWriteBlock == NULL ) {
// No write block set, find it!
    pWriteBlock = pHoldingRegs;
    while( pWriteBlock != NULL ) {
      if ( uintAddress >= pWriteBlock->uintAddress
        && uintAddress <= (pWriteBlock->uintAddress + 
                           pWriteBlock->uintTotal - 1)  ) {
        break;
      }
      pWriteBlock = pWriteBlock->pNext;
    }
  }
  if ( pWriteBlock != NULL ) {
    uint uintTemp, uintOffset;
    Lo(uintTemp) = bytDataLo;
    Hi(uintTemp) = bytDataHi;
    uintOffset = uintAddress - pWriteBlock->uintAddress;
    ((uint*)pWriteBlock->paryData)[uintOffset] = uintTemp;
    return TRUE;
  }
  return FALSE;
}
/**
 * Function:
 *  startTimeout
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
static void startTimeout(void) {
// Disable interrupt
  TMR0ON_bit = 0;
// Set timer presets
  if ( bytRxphase.B1 ) {
    TMR0H = Hi(uintRxGapSetPt3);
    TMR0L = Lo(uintRxGapSetPt3);
  } else if ( bytRxphase.B2 ) {
    TMR0H = Hi(uintRxGapSetPt2);
    TMR0L = Lo(uintRxGapSetPt2);
  } else if ( bytRxphase.B7 ) {
    TMR0H = Hi(uintRxGapSetPt1);
    TMR0L = Lo(uintRxGapSetPt1);
  } else {
    TMR0H = Hi(uintRxGapSetPt);
    TMR0L = Lo(uintRxGapSetPt);
  }
// Clear the interrupt mask
  TMR0IF_bit = 0;
// Enable interrupt
  TMR0ON_bit = 1;
}
/**
 * Function:
 *  restartRx
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
static void restartRx(void) {
// Clear FIFO
  bytRxphase = RCREG;
  bytRxphase = RCREG;
// Clear errors and enable Rx
  CREN_bit = 0;
  CREN_bit = 1;
// Set RS-485 direction
  Tx_dir = 0;
// Set parity check
  //RX9_bit = 1;
// ...or double STOP bit check
  //RX9D_bit = 1;
// Clear counters
  bytRxphase = 0;
  bytIndex = 0;
// Disable Tx
  TXEN_bit = 0;
  TXIE_bit = 0;
// Enable Rx interrupts
  RCIE_bit = 1;
  startTimeout();
}
/**
 * Function:
 *  addModbusBlock
 *
 * Parameters:
 *  eType, see modbusSlave.h mbType for options
 *  pBlock, the block to set-up
 *  uintAddress, the start address base 1
 *  uintTotal, the total number of I/O in the block
 *  paryData, a pointer to the actual data array,
 *              for discretes this must be an array of bytes
 *              for registers this must be an array of uints
 *  pCallback, optional call-back routine called when block is updated
 *
 * Returns:
 *  TRUE if block added, FALSE if not
 */
boolean addModbusBlock(mbType eType,
                       modbusBlockDef* pBlock,
                       uint uintAddress,
                       uint uintTotal,
                       void* paryData,
                       void (*pCallback)(struct _modbusBlock* pBlock)) {
  modbusBlockDef** pLL;
// Validate parameters
  if ( pBlock == NULL
    || uintAddress == 0
    || uintTotal == 0
    || paryData == NULL ) {
    return FALSE;
  }
  if ( eType == COILS ) {
    pLL   = &pCoils;
  } else if ( eType == STATUS_INPUTS ) {
    pLL   = &pStatusBits;
  } else if ( eType == HOLDING_REGISTERS ) {
    pLL   = &pHoldingRegs;
  } else if ( eType == INPUT_REGISTERS ) {
    pLL   = &pInputRegs;
  } else {
    return FALSE;
  }
  if ( *pLL == NULL ) {
// This is the first block
    *pLL = pBlock;
  } else {
// Make sure there are no blocks with the same address and type
    modbusBlockDef* pNode;
    for( pNode=*pLL; pNode->pNext!=NULL; pNode=pNode->pNext ) {
      if ( uintAddress >= pNode->uintAddress
        && uintAddress <= (pNode->uintAddress + pNode->uintTotal - 1)  ) {
// A block already exists using this address and for this type
        return FALSE;
      }
    }
// Append this block to the linked list
    pNode->pNext = pBlock;
  }
// Populate the block
  pBlock->eType       = eType;
  pBlock->uintAddress = uintAddress;
  pBlock->uintTotal   = uintTotal;
  pBlock->paryData    = paryData;
  pBlock->pCallback   = pCallback;
  pBlock->pNext       = NULL;
}
/**
 * Function:
 *  decodePacket
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
#pragma funcall decodePacket dummy

void decodePacket(void) {
  if ( TMR0IF_bit ) {
    bytRxphase.B0 = 1;
    TMR0ON_bit = 0;
// Clear the interrupt mask
    TMR0IF_bit = 0;

    if ( bytRxphase.B7 ) {
// Last byte sent, restart reception & set silent interval
      restartRx();
      return;
    }
  }
  if( RCIF_bit ) {
    byte bytTemp;
// Overrun or framing error?
    if ( OERR_bit || FERR_bit ) {
      restartRx();
      return;
    }
// May need to store RX9D_bit here for parity check!
// Read received byte
    bytTemp = RCREG;
// Silent interval or inter-char interval passed?
// If passed, is there enough space in buffer?
    if ( bytRxphase == 0 || bytRxphase.B2 == 1 ||
         bytIndex >= MAX_PACKET_LENGTH ) {
// No, wait for end of current message
      bytIndex = 0;
      bytRxphase = 0;
      startTimeout();  // 4.5 chars
      return;
    }
    if ( bytIndex == 0) {
// Address received
      bytRxphase = 2;
    }
// Store the data in buffer
    aryBuffer[bytIndex++] = bytTemp;
// Set inter-char interval
    startTimeout();  // 2.5 chars
  } else if ( bytRxphase.B7 ) {
    if ( TXIF_bit ) {
      if ( bytIndex < aryBuffer[0] ) {
        TXREG=aryBuffer[bytIndex++];
      } else {
// Wait for last byte to be transmitted
        startTimeout();  // 1 char
// Disable Tx interrupts
        TXIE_bit = 0;
      }
      return;
    }
  }
  if ( bytRxphase == 3 ) {
// Inter-char interval passed, check inter-frame timeout remainder
    bytRxphase = 4;
    startTimeout(); // 2 chars
  }
  if ( bytRxphase != 5 ) {
// Nothing more to do
    return;
  }
// Message received, our addreess?
  if (aryBuffer[0] == bytSlaveAddress ) {
    uint uintCRCpos, uintCRC;
// Set the expected position of the CRC
    uintCRCpos = 6;
// Is the function supported?
    switch( aryBuffer[1] ) {
    case READ_COILS:
    case FORCE_SINGLE_COIL:
    case FORCE_MULTIPLE_COILS:
      if ( pCoils == NULL ) {
// No coils defined!
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      }
      break;
    case READ_STATUS_INPUTS:
      if ( pStatusBits == NULL ) {
// No status inputs defined!
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      }
      break;
    case READ_HOLDING_REGISTERS:
    case PRESET_SINGLE_REGISTER:
    case PRESET_MULTIPLE_REGISTERS:
      if ( pHoldingRegs == NULL ) {
// No holding registers defined!
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      }
      break;
    case READ_INPUT_REGISTERS:
      if ( pInputRegs == NULL ) {
// No input registers defined!
        exceptionCode = ILLEGAL_DATA_ADDRESS;
      }
      break;
    default:
// No, function not supported!
      exceptionCode = ILLEGAL_FUNCTION;
      break;
    }
    if ( aryBuffer[1] == FORCE_MULTIPLE_COILS ||
         aryBuffer[1] == PRESET_MULTIPLE_REGISTERS )  {
      uintCRCpos += 1 + aryBuffer[6];
    }
// Calculate the packet CRC
    uintCRC = calcCRC(uintCRCpos);
// Compare the calculate checksum with the received checksum
    if ( Lo(uintCRC)  == aryBuffer[uintCRCpos] &&
         Hi(uintCRC) == aryBuffer[uintCRCpos + 1] ) {
// CRC ok
      uint uintSAddr, uintEAddr, uintItemCount, uintLength = 6;
      uint uintOffset, uintAddr;
      boolean blnState;
      mbType eType;
      byte bytBit;
// Get the start address
      Lo(uintSAddr) = aryBuffer[3];
      Hi(uintSAddr) = aryBuffer[2];
// How many items have been requested?
      Lo(uintItemCount) = aryBuffer[5];
      Hi(uintItemCount) = aryBuffer[4];
// Whats the last address?
      uintEAddr = uintSAddr + uintItemCount;
      uintSAddr++;
// Make sure write block is not set
      pWriteBlock = NULL;
// What was the requested function code?
      switch( aryBuffer[1] ) {
      case READ_COILS:
      case READ_STATUS_INPUTS:
        if ( aryBuffer[1] == READ_COILS ) {
          eType = COILS;
        } else {
          eType = STATUS_INPUTS;
        }
        aryBuffer[2] = packBits(eType, &aryBuffer[3], uintSAddr, uintEAddr);
        
        if ( aryBuffer[2] == 0 ) {
// Exception, address does not exist
          exceptionCode = ILLEGAL_DATA_ADDRESS;
        } else {
          uintLength = aryBuffer[2] + 3;
        }
        break;
      case READ_HOLDING_REGISTERS:
      case READ_INPUT_REGISTERS:
        if ( aryBuffer[1] == READ_INPUT_REGISTERS ) {
          eType = INPUT_REGISTERS;
        } else {
          eType = HOLDING_REGISTERS;
        }
        aryBuffer[2] = packRegisters(eType, &aryBuffer[3], 
                                     uintSAddr, uintEAddr);
        if ( aryBuffer[2] == 0 ) {
// Exception, address does not exist
          exceptionCode = ILLEGAL_DATA_ADDRESS;
        } else {
          uintLength = aryBuffer[2] + 3;
        }
        break;
      case FORCE_SINGLE_COIL:
        if ( aryBuffer[4] == 0xff && aryBuffer[5] == 0x0 ) {
// Force coil on
          blnState = coilState(uintSAddr, TRUE);
        } else if ( aryBuffer[4] == 0x0 && aryBuffer[5] == 0x0 ) {
// Force coil off
          blnState = coilState(uintSAddr, FALSE);
        }
        if ( blnState == FALSE ) {
// Exception, address does not exist
          exceptionCode = ILLEGAL_DATA_ADDRESS;
        }
        break;
      case PRESET_SINGLE_REGISTER:
        if ( setRegister(uintSAddr, aryBuffer[4], aryBuffer[5]) == FALSE ) {
// Exception, address does not exist
          exceptionCode = ILLEGAL_DATA_ADDRESS;
        }
        break;
      case FORCE_MULTIPLE_COILS:
        if ( uintItemCount > MAX_DISCRETES_IN_FC15 ||
             uintItemCount > (aryBuffer[6] * 8) ) {
// Exception, either to many items or the coil count doesn't match byte count
          exceptionCode = ILLEGAL_DATA_ADDRESS;
          break;
        }
      case PRESET_MULTIPLE_REGISTERS:
        if ( (aryBuffer[1] == PRESET_MULTIPLE_REGISTERS &&
             uintItemCount > MAX_REGISTERS_IN_FC16) ) {
// Exception, either to many items or the coil count doesn't match byte count
          exceptionCode = ILLEGAL_DATA_ADDRESS;
        } else {
          bytBit = 1;
          uintOffset = 0;
          uintAddr = uintSAddr;

          while( uintAddr <= uintEAddr ) {
            if ( aryBuffer[1] == PRESET_MULTIPLE_REGISTERS ) {
              if ( setRegister(uintAddr,
                               aryBuffer[7 + uintOffset],
                               aryBuffer[8 + uintOffset]) == FALSE ) {
                exceptionCode = ILLEGAL_DATA_ADDRESS;
                break;
              }
              uintOffset += 2;
              uintAddr++;
            } else {
              blnState = FALSE;
              if ( (aryBuffer[7 + uintOffset] & bytBit) > 0 ) {
                blnState = TRUE;
              }
              if ( coilState(uintAddr, blnState) == FALSE ) {
                exceptionCode = ILLEGAL_DATA_ADDRESS;
                break;
              }
              bytBit <<= 1;
              if ( bytBit == 0 ) {
                bytBit = 1;
                uintOffset++;
              }
              uintAddr++;
            }
          }
        }
        break;
      }
      if( exceptionCode != NO_EXCEPTION ) {
        aryBuffer[1] |= EXCEPTION_FLAG;
        aryBuffer[2] = (byte)exceptionCode;
        exceptionCode = NO_EXCEPTION;
        uintLength = 3;
      }
      if ( uintLength > 0 ) {
// A write block has been set, if it has a callback, then call it!
       if ( pWriteBlock != NULL && pWriteBlock->pCallback != NULL ) {
         (*pWriteBlock->pCallback)(pWriteBlock);
       }
// Calculate CRC for response
        uintCRC = calcCRC(uintLength);
        aryBuffer[uintLength++] = Lo(uintCRC);
        aryBuffer[uintLength++] = Hi(uintCRC);
// Stop receiving
        CREN_bit = 0;
        RCIE_bit = 0;
// Enable transmission, set direction
        TXEN_bit = 1;
        Tx_dir = 1;
// Mark start of transmission
        bytRxphase = 128;
// Send address
        TXREG = bytSlaveAddress;
// Store length and set index
        aryBuffer[0] = uintLength;
        bytIndex = 2;
// Send function code to fill Tx buffer and avoid instant return to ISR
        while( !TXIF_bit ){}
        TXREG = aryBuffer[1];
// Enable Tx interrupts for remainder of response to follow
        TXIE_bit = 1;
        return;
      }
    }
  }
// No need to answer (not our address or message frame error, await next one)
  bytIndex = 0;
  bytRxphase = 1;
}
/**
 * Function:
 * modbusSlave
 *
 * Parameters:
 *  eBaud, see modbusSlave.h for options
 *  bytAddress, the modbus slave address 1 to 255
 *  bytStopBits, number of transmitted stop bits, 1 or 2
 *
 * Returns:
 *  0 if ok, -1 if error
 */
int modbusSlave(baudRate eBaud, const byte bytAddress, const byte bytStopBits) {
  long lngTimeoutPreset;
  byte bytTemp;

  if ( bytAddress < 1 || bytAddress > 247 ) {
    return -1;
  }
// Make sure the received array is cleared
  memset(aryBuffer, 0, sizeof(aryBuffer));
// Set the slave address
  bytSlaveAddress = bytAddress;
// Initialise the Rx Gap counter
  uintRxGap = 0;
// Ensure there is no exception code
  exceptionCode = NO_EXCEPTION;
// Initialize UART module for the specified bps
  switch( eBaud ) {
  case BAUD_1200:
    UART1_Init(BAUD_1200 * 100);
    lngTimeoutPreset = GAP_SETPT_1200;
    break;
  case BAUD_2400:
    UART1_Init(BAUD_2400 * 100);
    lngTimeoutPreset = GAP_SETPT_2400;
    break;
  case BAUD_4800:
    UART1_Init(BAUD_4800 * 100);
    lngTimeoutPreset = GAP_SETPT_4800;
    break;
  case BAUD_9600:
    UART1_Init(BAUD_9600 * 100);
    lngTimeoutPreset = GAP_SETPT_9600;
    break;
  case BAUD_19200:
    UART1_Init(BAUD_19200 * 100);
    lngTimeoutPreset = GAP_SETPT_19200;
    break;
 case BAUD_38400:
    UART1_Init(BAUD_38400 * 100);
    lngTimeoutPreset = GAP_SETPT_38400;
    break;
  case BAUD_57600:
    UART1_Init(BAUD_57600 * 100);
    lngTimeoutPreset = GAP_SETPT_57600;
    break;
  case BAUD_115200:
    UART1_Init(BAUD_115200 * 100);
    lngTimeoutPreset = GAP_SETPT_115200;
    break;
  default:
    return -1;
  }
// Wait for UART module to stabilize
  Delay_ms(100);
// Timer0 Registers:
// 16-Bit Mode
// Prescaler=1:1
  T0CON.TMR0ON = 0;// Timer0 On/Off Control bit 1=Enables / 0=Stops
  T0CON.T08BIT = 0;// Timer0 8-bit/16-bit Control bit: 1=8-bit/ 0=16-bit
  T0CON.T0CS   = 0;// TMR0 Clock Source Select bit: 0=Internal
  T0CON.T0SE   = 0;// TMR0 Source Edge Select bit: 0=low/high
  T0CON.PSA    = 1;// Prescaler Assignment bit: 1=NOT assigned/bypassed
  T0CON.T0PS2  = 0;// bits 2-0  PS2:PS0: Prescaler Select bits
  T0CON.T0PS1  = 0;
  T0CON.T0PS0  = 0;
// Set prescaler if needed
  bytTemp = 0;
  while( lngTimeoutPreset > 29126 ) {
    lngTimeoutPreset >>= 1;
    bytTemp++;
  }
  if ( bytTemp > 0 && bytTemp <= 8 ) {
    T0CON.PSA = 0;   // Prescaler assigned
    T0CON = T0CON | (--bytTemp & 0x07);
  }
  uintRxGapSetPt1 = (uint)-(lngTimeoutPreset >> 1);   // 1 char
  uintRxGapSetPt2 = (uint)-lngTimeoutPreset;          // 2 chars
  uintRxGapSetPt3 = (uint)-(lngTimeoutPreset*5 >> 2); // 2.5 chars
  uintRxGapSetPt  = (uint)-(lngTimeoutPreset*9 >> 2); // 4.5 chars
// Double STOP bit while transmitting?
  if ( bytStopBits != 1 ) {
    TX9_bit  = 1;
    TX9D_bit = 1;
  }
// Clear errors, Rx buffer & Rx phase indicator
// and start silent interval in case bus is active
  restartRx();
  TMR0IE_bit = 1;
// Enable peripheral interrupts
  PEIE_bit   = 1;
// Enable GLOBAL interrupts
  GIE_bit    = 1;
}